@@ -3,48 +3,42 @@ package Debug::Filter::PrintExpr;
 use strict;
 use warnings;
  
-use Exporter::Tiny;
-
+use 5.010;
 use Filter::Simple;
 use Scalar::Util qw(isdual blessed);
+use List::Util 'pairs';
 use Data::Dumper;
+use B qw(svref_2object SVf_IOK SVp_IOK SVf_NOK SVp_NOK);
 
 our
-$VERSION = '0.10';
-
-our @EXPORT_OK = qw(isnumeric isstring);
-our @ISA = qw(Exporter::Tiny);
-our %EXPORT_TAGS = (
-	debug => [],
-	nofilter => [],
-	all	=> [qw(isnumeric isstring)],
-);
-
-require XSLoader;
-XSLoader::load('Debug::Filter::PrintExpr', $VERSION);
-
-local ($,, $\);
-
-sub import {
-	goto \&Exporter::Tiny::import;
-}
+$VERSION = '0.19';
 
 # variable is exposed and my be overwritten by caller
 our $handle = *STDERR;
 
-# generate a prefix containing line number or custom label
-sub _genprefix {
-	my ($self, $label, $line, $expr, $value) = @_;
-	local ($,, $\);
-	print $handle $label ? $label : "line $line:";
-	print $handle " $expr = " if $expr;
+# helper function: tests if given variable has data
+# in it's "numeric slot"
+sub _isnumeric {
+	return svref_2object(\$_[0])->FLAGS & 
+		(SVf_IOK | SVp_IOK | SVf_NOK | SVp_NOK);
 }
 
+# generate a prefix containing line number or custom label
+# consume first three args, return number of printed chars
+# if no expression is present
+sub _genprefix {
+	my ($label, $line, $expr, $pos) = splice @_, 0, 3;
+	local ($,, $\);
+	printf $handle "%s%n", $label || "L$line:", $pos;
+	print $handle $expr ? " $expr = " : " ";
+	return $expr ? undef : $pos + 1;
+}
+
+# create representation of single value
 sub _singlevalue {
-	my $val = shift;
-	my ($str, $num);
+	my ($val, $str, $num) = shift;
 	my $isdual = isdual($val);
-	my $isnumeric = isnumeric($val);
+	my $isnumeric = _isnumeric($val);
 	$str = "$val" if defined $val;
 	$num = $val + 0 if $isnumeric;
 	if (!defined $val) {
@@ -54,7 +48,7 @@ sub _singlevalue {
 	} elsif (ref($val)) {
 		return $val;
 	} elsif ($isdual) {
-		return "'$str' : $num";
+		return "dualvar($num, '$str')";
 	} elsif ($isnumeric) {
 		return $num;
 	} else {
@@ -71,157 +65,102 @@ sub _valuescalar {
 # print out an expression in list context
 sub _valuearray {
 	local ($,, $\);
-	print $handle join(', ', map({_singlevalue($_)} @_));
+	print $handle '(', join(', ',
+		map({_singlevalue($_)} @_)), ");\n";
 }
 
-# print out an key-value pair, prefixed by a seperator
+# print out an expression as key-value pairs
 sub _valuehash {
-	our $sep;
 	local ($,, $\);
-	local *sep = \$_[0];
-	shift;
-	my ($key, $value) = @_;
-	print $handle $sep || '',
-		"'$key' => ", _singlevalue($value);
-	$sep = ', ';
+	print $handle '(', join(", ",
+		map(
+		{"'$_->[0]' => " . _singlevalue($_->[1])}
+		pairs(@_))), ");\n";
 }
 
-# print opening before list/hash data
-sub _genopen {
+# process a scalar debug statement
+sub _print_scalar {
 	local ($,, $\);
-	print $handle '(';
-}
-
-# print closing after list/hash data
-sub _genclose {
-	local ($,, $\);
-	print $handle ");\n";
-}
-
-# process a complete scalar debug statement
-sub _printscalar {
-	my ($self, $label, $line, $expr, $value) = @_;
-	local ($,, $\);
-	_genprefix @_;
-	_valuescalar($_[4]) if $expr;
-	print $handle ';' if $expr;
+	unless (&_genprefix) {
+		_valuescalar($_[0]);
+		print $handle ';';
+	}
 	print $handle "\n";
 }
 
-sub _printstr {
-	my ($self, $label, $line, $expr, $value) = @_;
-	_printscalar($self, $label, $line, $expr, "$value");
+# process a string scalar debug statement
+sub _print_str {
+	my $val = $_[3];
+	splice @_, 3, 1, "$val";
+	goto &_print_scalar;
 }
 
-sub _printnum {
+# process a numeric scalar debug statement
+sub _print_num {
 	no warnings qw(numeric);
-	my ($self, $label, $line, $expr, $value) = @_;
-	_printscalar($self, $label, $line, $expr, $value+0);
+	my $val = $_[3];
+	splice @_, 3, 1, $val + 0;
+	goto &_print_scalar;
 }
 
-
-# process a complete array debug statement
-sub _printarray {
-	my ($self, $label, $line, $expr, @value) = @_;
-	_genprefix @_;
-	_genopen;
-	_valuearray @value;
-	_genclose;
+# process an array debug statement
+sub _print_array {
+	&_genprefix;
+	goto &_valuearray;
 }
 
-# start a hash debug statement
-sub _printhashopen {
-	my ($self, $label, $line, $expr) = @_;
-	_genprefix @_;
-	_genopen;
+# process a hash debug statement
+sub _print_hash {
+	&_genprefix;
+	goto &_valuehash;
 }
 
-# process a single iteration of a key-value pair
-sub _printhashitem {
-	my ($sep, $key, $value) = @_;
-	_valuehash @_;
-}
-
-# end a hash debug statement
-sub _printhashclose {
-	_genclose;
-}
-
-# process a complete reference debug statement
-sub _printref {
-	my ($self, $label, $line, $expr, @value) = @_;
+# process a reference debug statement
+sub _print_ref {
+	my $expr = splice @_, 2, 1, undef;
+	my $skip = &_genprefix;
 	local ($,, $\);
-	print $handle $label ? $label : "line $line:", " ";
-	my $d = Data::Dumper->new([@value]);
-	my $dump =  $d->Dump;
-	if (scalar(@value) <= 1) {
-		$dump =~ s/^\$VAR1/$expr/;
-	} else {
-		$dump =~ s/^\$VAR(\d+)/"($expr)[" . ($1 - 1) . "]"/meg;
-	}
-	print $handle "\n", $dump;
+	print $handle "dump($expr);\n";
+	print $handle 
+		Data::Dumper->new([@_], [map("_[$_]", (0 .. $#_))])
+		->Pad(' ' x $skip)->Dump;
 }
 
-# process a debug statement
+# type classifications: print function suffix + is scalar
+my %type_defs = (
+	'$' => ['scalar', 1],
+	'"' => ['str', 1],
+	'#' => ['num', 1],
+	'@' => ['array', 0],
+	'%' => ['hash', 0],
+	'\\' => ['ref', 0],
+);
+
+# process a debug statement, runs in filter context
 sub _gen_print {
-	my ($self, $type, $label, $expr) = @_;
-	$label ||= '';
-	if ($type eq '$') {
-		my $print = $self . "::_printscalar";
-		$expr ||= '';
-		return qq[{$print("$self", "$label", __LINE__, q{$expr}, scalar(($expr)));}];
-	} elsif ($type eq '"') {
-		my $print = $self . "::_printstr";
-		$expr ||= '';
-		return qq[{$print("$self", "$label", __LINE__, q{$expr}, scalar($expr));}];
-	} elsif ($type eq '#') {
-		my $print = $self . "::_printnum";
-		$expr ||= '';
-		return qq[{$print("$self", "$label", __LINE__, q{$expr}, scalar($expr));}];
-	} elsif ($type eq '@') {
-		my $print = $self . "::_printarray";
-		return qq[{$print("$self", "$label", __LINE__, q{$expr}, $expr);}];
-	} elsif ($type eq '%') {
-		my $printopen = $self . '::_printhashopen';
-		my $printitem = $self . '::_printhashitem';
-		my $printclose = $self . '::_printhashclose';
-		my $sep = '$' . $self . '::sep';
-		my $pair = '@' . $self . '::pair';
-		my $stmt = qq[{local ($sep, $pair); ];
-		$stmt .= qq[$printopen("$self", "$label", __LINE__, q{$expr}); ];
-		$stmt .= qq[$printitem($sep, $pair) ];
-		$stmt .= qq[while $pair = each($expr); ];
-		$stmt .= qq[$printclose;}];
-		return $stmt;
-	} elsif ($type eq '\\') {
-		my $print = $self . "::_printref";
-		return qq[{$print("$self", "$label", __LINE__, q{$expr}, $expr);}];
-	} else {
-		return '# type unknown';
-	}
+	my ($type, $label, $expr) = map $_ // '', @_;
+	my $val = $_[2] // '()';
+	my ($ptype, $scalar) = @{$type_defs{$type}};
+	my $print = __PACKAGE__ . "::_print_$ptype";
+	return qq[{$print("$label", __LINE__, q{$expr}, ] .
+		($scalar ? qq[scalar($val)] : qq[$val]) .
+		q[)}];
 }
 
 # source code processing happens here
 FILTER {
 	my ($self, @args) = @_;
-	my ($nofilter, $debug);
-	if (ref($_[1]) eq 'HASH') {
-		my $global = $_[1];
-		$debug = $global->{debug};
-		$nofilter = $global->{nofilter};
-	}
+	my $debug;
 	$debug ||= grep /^-debug$/, @args;
-	$nofilter ||= grep /^-nofilter$/, @args;
 	s/
-		^\h*\#
+		^\h*+\#
 		(?<type>[%@\$\\#"])
-		\{\h*
-		(?<label>[[:alpha:]_]\w*:)?
-		\h*
-		(?<expr>\V+)?
-		\}\h*\r?$
-	/ _gen_print($self, $+{type}, $+{label}, $+{expr}) /gmex
-		unless $nofilter;
+		\{\h*+
+		(?<label>[[:alpha:]_]\w*+:)?
+		\h*+
+		(?<expr>\V*[^\s])?\h*
+		\}\h*+\r?$
+	/ _gen_print($+{type}, $+{label}, $+{expr}) /gmex;
 	print STDERR if $debug;
 };
 
@@ -251,19 +190,19 @@ Debug::Filter::PrintExpr - Convert comment lines to debug print statements
 	#${ calc: @a * 2 }
 	#\{$ref}
 
-This program produces an output like this:
+This produces an output like:
 
-	line 13: $s = 'a scalar';
-	line 14: @a = ('this', 'is', 'an', 'array');
-	line 15: %h = ('' => 'empty', 'key1' => 'value1', 'key2' => 'value2', 'undef' => undef);
-	calc: @a * 2  = 8;
-	line 17: 
-	$ref = {
-		  '' => 'empty',
-		  'key1' => 'value1',
-		  'key2' => 'value2',
-		  'undef' => undef
-		};
+	L13: $s = 'a scalar';
+	L14: @a = ('this', 'is', 'an', 'array');
+	L15: %h = ('' => 'empty', 'key1' => 'value1', 'key2' => 'value2', 'undef' => undef);
+	calc: @a * 2 = 8;
+	L17: dump($ref);
+	     $_[0] = {
+	               '' => 'empty',
+	               'key1' => 'value1',
+	               'key2' => 'value2',
+	               'undef' => undef
+	             };
 
 =head1 DESCRIPTION
 
@@ -292,11 +231,16 @@ for various debugging purposes.
 (Besides, there is L<Smart::Comments> from Damian, that does something
 very similar but more advanced.)
 
-Just by removing the "use" of C<Debug::Filter::PrintExpr> completely
-or disabling it partially by
+Just by removing the "use" of C<Debug::Filter::PrintExpr> completely,
+disabling it partially by
 
 	no Debug::Filter::PrintExpr;
 
+or making the usage conditional (e.g. on environment variable DEBUG)
+by
+
+	use if $ENV{DEBUG}, 'Debug::Filter::PrintExpr';
+
 all these lines (or a part of them) lose their magic and remain
 simple comments.
 
@@ -319,7 +263,7 @@ or more formally must be matched by the following regexp:
 where C<type> represents the sigil, C<label> an optional label and
 C<expr> an optional expression.
 
-If the label is omitted, it defaults to C<line nnn:>, where nnn is the
+If the label is omitted, it defaults to C<LI<n>:>, where n is the
 line number in the program.
 
 The sigil determines the evaluation context for the given expression
@@ -331,30 +275,13 @@ and the output format of the result:
 
 The expression is evaluated in scalar context. Strings are printed
 inside single quotes, integer and floating point numbers are
-printed unquoted and dual valued variables are shown in both
-representations seperated by a colon.
+printed unquoted and dual valued variables are shown in the form
+C<dualvar(I<numval>, 'I<stringval>')>.
 Undefined values are represented by the unquoted string C<undef>.
 Hash and array references are shown in their usual string representation
 as e.g. C<ARRAY(0x19830d0)> or C<HASH(0xccba88)>.
 Blessed references are shown by the class they are belong to as
-C<blessed(I<class>)>.
-
-=item C<@>
-
-The expression is evaluated in list context and the elements of the
-list are printed like single scalars, separated by commas and gathered
-in parentheses.
-
-=item C<%>
-
-The expression is used as argument in a while-each loop and the output
-consists of pairs of the form 'key' => I<value> inside parentheses.
-I<value> is formatted like a single scalar.
-
-=item C<\>
-
-The expression shall be a list of references.
-These will be evaluated using L<Data::Dumper>.
+C<< blessed(class) >>.
 
 =item C<">
 
@@ -364,21 +291,39 @@ The expression is evaluated in scalar context as a string.
 
 The expression is evaluated in scalar context as a numeric value.
 
+=item C<@>
+
+The expression is evaluated in list context and the elements of the
+list are printed like single scalars, separated by commas and gathered
+in parentheses.
+
+=item C<%>
+
+The expression is evaluated as a list of key-value pairs
+and is presented in the form 'key' => I<value>,... inside parentheses.
+I<value> is formatted like a single scalar.
+
+=item C<\>
+
+The expression shall evaluate to a list of references.
+These will be evaluated using L<Data::Dumper> and named
+like parameters in a subroutine, i.e. C<$_[I<n>]>.
+
 =back
 
 The usage and difference between C<#${}>, C<#"{}> and C<##{}> is
 best described by example:
 
 	my $dt = DateTime->now;
-	#${$dt}		# line nn: $dt = blessed(DateTime);
-	#"{$dt}		# line nn: $dt = '2019-10-27T15:54:28';
+	#${$dt}		# Ln: $dt = blessed(DateTime);
+	#"{$dt}		# Ln: $dt = '2019-10-27T15:54:28';
 
 	my $num = ' 42 ';
-	#${$num}	# line nn: $num = ' 42 ';
+	#${$num}	# Ln: $num = ' 42 ';
 	$num + 0;
-	#${$num}	# line nn: $num = ' 42 ' : 42;
-	#"{$num}	# line nn: $num = ' 42 ';
-	##{$num}	# line nn: $num = 42;
+	#${$num}	# Ln: $num = dualvar(42, ' 42 ');
+	#"{$num}	# Ln: $num = ' 42 ';
+	##{$num}	# Ln: $num = 42;
 
 
 The forms #${}, #"{}, ##{} and #@{} may be used for any type of expression
@@ -389,14 +334,12 @@ to use:
 	#@{scalar_as_array: $s}
 	#${array_as_scalar :@a}
 	#@{hash_as_array: %h}
-	#%{array_as_hash: @a}
 
 and produce these results:
 
 	scalar_as_array: $s = ('this is a scalar');
 	array_as_scalar: @a = 4;
 	hash_as_array: %h = ('k1', 'v1', 'k2', 'v2');
-	array_as_hash: @a = ('0' => 'this', '1' => 'is', '2' => 'an', '3' => 'array');
 	
 Regular expressions may be evaluated too:
 
@@ -404,7 +347,7 @@ Regular expressions may be evaluated too:
 
 gives:
 
-	line nn: "a<b>c<d><e>f<g>h" =~ /\w*<(\w+)>/g = ('b', 'd', 'e', 'g');
+	Ln: "a<b>c<d><e>f<g>h" =~ /\w*<(\w+)>/g = ('b', 'd', 'e', 'g');
 
 If the expression is omitted, only the label will be printed.
 The sigil C<$> should be used in this case.
@@ -419,8 +362,8 @@ It must be a valid Perl expression.
 
 =item *
 
-In case of the #%{}-form, it must be a valid argument to the
-each() builtin function, i.e. it should resolve to an array or hash.
+In case of the #%{}-form, it must evaluate to a list of pairs, e.g.
+a hash.
 
 =back
 
@@ -440,56 +383,14 @@ about a useless use of something in void context.
 =head2 Usage
 
 The C<use> statement for C<Debug::Filter::PrintExpr> may contain
-arguments as described in L<Exporter::Tiny::Manual::Importing>.
-Importable functions are C<isnumeric> and C<isstring> as well
-as the import tag C<:all> for both of them.
-
-The (optional) global options hash may contain
-these module specific entries:
+following arguments:
 
 =over 4
 
-=item debug => 1
+=item -debug
 
 This option causes the resulting source code after comment
 transformation to be written to C<STDERR>.
-This option may also be specified as C<-debug> in the
-C<use> statement.
-
-=item nofilter => 1
-
-This options disables source code filtering if only the import
-of functions is desired.
-This option may also be specified as C<-nofilter> in the
-C<use> statement.
-
-=back
-
-=head2 Functions
-
-=over
-
-=item C<isstring(I<$var>)>
-
-This function returns true if the "string slot" of I<$var> has a value.
-This is the case when a string value was assigned to the variable,
-the variable has been used (recently) in a string context
-or when the variable is dual-valued.
-
-It will return false for undefined variables, references and
-variables with a numeric value that have never been used in a
-string context.
-
-=item C<isnumeric(I<$var>)>
-
-This function returns true if the "numeric slot" if I<$var> has a
-value.
-This is the case when a numeric value (integer or floating point) was
-assigned to the variable, the variable has been used (recently) in a
-numeric context or when the variable is dual-valued.
-
-It will return false for undefined variables, references and variables
-with a string value that have never been used in numeric context.
 
 =back
 
@@ -505,7 +406,6 @@ The default is STDERR and may be changed by the caller.
 
 =back
 
-
 =head1 SEE ALSO
 
 Damian Conway's module L<Smart::Comments> provides something similar
@@ -529,7 +429,7 @@ Always print the literal expression along with its evaluation.
 
 Give a defined context where the expression is evaluated.
 Especially provide scalar and list context or perform an iteration
-over a while-each-loop.
+over the key-value pairs of a hash.
 The usage of L<Data::Dumper> was adopted later from Damian's
 implementation.
 
@@ -555,10 +455,6 @@ and a specific context is not enforced by the module.
 All in all, the module presented here is not much more than a
 programming exercise.
 
-Importing the functions C<isstring> and C<isnumeric> is done
-by L<Exporter::Tiny>.
-For extended options see L<Exporter::Tiny::Manual::Importing>.
-
 Other related modules: L<Scalar::Util>, L<Data::Dumper>
 
 =head1 AUTHOR
@@ -567,7 +463,7 @@ Jörg Sommrey
 
 =head1 LICENCE AND COPYRIGHT
 
-Copyright (c) 2018-2019, Jörg Sommrey. All rights reserved.
+Copyright (c) 2018-2020, Jörg Sommrey. All rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
@@ -575,5 +471,4 @@ by the Free Software Foundation; or the Artistic License.
 
 See L<http://dev.perl.org/licenses/> for more information.
 
-
 =cut
